# version: 2
from __future__ import annotations
import abc
from functools import partial
from math import floor
from multiprocess.pool import Pool   # Uses dill for serialising.
from multiprocessing import shared_memory
import os
import re
import shutil
from typing import List, Tuple, Union

import numpy as np
import pandas as pd

from expam.classify.config import create_tax_results, load_results_config
from expam.classify.process import ExpamClassifierProcesses
from expam.classify.taxonomy import TaxonomyInterface
from expam.database import EXPAM_TEMP_EXT, TIMEOUT, UNION_RATIO
from expam.database.config import load_database_config
from expam.database.db import SharedDB
from expam.process.classify import ClassifyWorker
from expam.process.manager import ControlCenter
from expam.process.reads import ReadWorker
from expam.tree.tree import Index
from expam.utils import abspath, die, ls


def run_classifier(read_paths: str, out_dir: str, db_dir: str, k: int, n: int, phylogeny_path: str,
                   keys_shape: Tuple[int], values_shape: Tuple[int], logging_dir: str,
                   taxonomy: bool = False, groups: Union[None, List[tuple]] = None,
                   keep_zeros: bool = False, cpm: float = 0.0, use_node_names: bool = True,
                   colour_list: List[str] = None, paired_end: bool = False, alpha: float = 1.0,
                   log_scores: bool = False, itol_mode: bool = False, flat_colour: bool = False,
                   sep: str = ","):
    database_config = load_database_config(db_dir)
    output_config = load_results_config(out_dir, create=True)
    # Load the kmer dictionary.
    print("Loading the map and phylogeny.\n")
    db_kmers = SharedDB(db_dir, keys_shape, values_shape)
    try:
        # Get name mapping.
        _, tree = Index.load_newick(phylogeny_path)
        # Load LCA Matrix.
        lca_matrix = np.load(database_config.lca_matrix)
        # Perform classification and phylogenetic summary.
        c = ClassificationRun(k, db_kmers, tree, lca_matrix, read_paths, out_dir,
                              output_config.temp, logging_dir, alpha,
                              keep_zeros=keep_zeros, cpm=cpm,
                              paired_end=paired_end, sep=sep, n_processes=n)
        phy = c.run()
        # Taxonomic results.
        if taxonomy:
            create_tax_results(output_config)
            my_ncbi = TaxonomyInterface(database_config, tree=tree)
            tax = TaxonomicResults.from_phylogenetic(phy, my_ncbi)
            tax.summarise(output_config.tax,
                          cpm=cpm,
                          sep=sep)
            tax.convert_raw(output_config.phy_raw, output_config.tax_raw, tree)
        # Plot phylogenetic results.
        phy.draw_results(output_config.phy, groups=groups, cpm=cpm,
                         colour_list=colour_list, keep_zeros=keep_zeros,
                         use_node_names=use_node_names, log_scores=log_scores,
                         itol_mode=itol_mode, flat_colour=flat_colour)
    finally:
        # Close shared memory.
        keys_shm = shared_memory.SharedMemory(name=db_kmers.keys_shm_name, create=False)
        values_shm = shared_memory.SharedMemory(name=db_kmers.values_shm_name, create=False)
        keys_shm.unlink()
        keys_shm.close()
        values_shm.unlink()
        values_shm.close()
        shutil.rmtree(output_config.temp)

class ClassificationRun:
    def __init__(self, k: int, kmer_db: SharedDB, tree: Index, lca_matrix: np.ndarray,
                 read_paths: List[str], out_dir: str, temp_dir: str, logging_dir: str, alpha: float,
                 keep_zeros: bool = False, cpm: float = 0.0, paired_end: bool = False,
                 sep: str = ",", n_processes: int = 1):
        """

        :param k: Int
        :param kmer_db: SharedDB
        :param lca_matrix: np.ndarray
        :param url: str
        :param out_dir: str
        """
        # Set up processing references.
        self.k = k
        self.alpha = alpha
        self.kmer_db = kmer_db
        self.lca_matrix = lca_matrix
        self.out_dir = out_dir
        self.temp_dir = temp_dir
        self.logging_dir = logging_dir

        self.tree = tree
        self.node_names: List[str] = [
            node.name if i > 0 else "unclassified"
            for i, node in enumerate(tree.pool)
        ]

        self.keep_zeros = keep_zeros
        self.cpm = cpm

        # Url containing reads.
        self.results_config = load_results_config(out_dir, create=True)
        self.read_paths = read_paths
        self.paired_end = paired_end

        self.SEP = sep
        self.n_processes = n_processes

    def run(self) -> PhylogeneticResults:
        """
        Begin processing reads in self.url.
        We have some collection of processes accessing reads from storage
        and then another collection of processes that analyse the reads.

        """
        ## sample processing
        #
        file_queue = self.prepare_queue(self.paired_end)
        if len(file_queue) == 0:
            die("Could not find any files in these directories: %s" % (", ".join(self.read_paths)))

        n_classifiers = floor(UNION_RATIO * self.n_processes)
        n_readers = self.n_processes - n_classifiers
        mp_config = {
            "name": "expam",
            "phases": ["classify"],
            "layers": {
                "class": ReadWorker,
                "class_args": {
                    "paired_mode": self.paired_end,
                    "n_classifiers": n_classifiers,
                    "logging_dir": self.logging_dir,
                },
                "parent": ControlCenter,
                "child": {
                    "class": ClassifyWorker,
                    "class_args": {
                        "k": self.k,
                        "kmer_db": self.kmer_db,
                        "lca_matrix": self.lca_matrix,
                        "out_dir": self.results_config.phy,
                        "temp_dir": self.temp_dir,
                        "n_readers": n_readers,
                        "logging_dir": self.logging_dir,
                        "alpha": self.alpha
                    },
                    "parent": ReadWorker,
                    "child": None,
                    "n": n_classifiers
                },
                "n": n_readers
            },
            "timeout": TIMEOUT
        }

        process_network = ExpamClassifierProcesses.from_method_dict(self.logging_dir, mp_config)
        # load jobs
        for file_dir in file_queue:
            process_network.load_job(phase="classify", job=file_dir)
        process_network.run()

        ## data formatting
        #
        # concatenate process-wise output
        self.combine_raw_read_output()
        # create phylogenetic output
        phy = PhylogeneticResults.from_raw_read_output(self.results_config.phy_raw,
                                                       num_procs=self.n_processes,
                                                       tree=self.tree)
        phy.summarise(self.results_config.phy,
                      cpm=self.cpm,
                      sep=self.SEP)
        phy.to_csv(self.results_config.phy_classified,
                   self.results_config.phy_split,
                   self.keep_zeros)
        return phy

    def prepare_queue(self, paired_end: bool = False) -> List[Tuple[str, ...]]:
        file_queue = []
        for path in self.read_paths:
            # Create a list of all the files containing reads to be processed.
            file_list = [
                abspath(file_name)
                for file_name in ls(path)
                if re.match(r"(\S+)\.(?:fa|fna|ffn|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_name)
            ]
            if paired_end:  # This implies path is a folder, not a file.
                while file_list:
                    next_file = file_list.pop()
                    # Find buddy.
                    match_ind = self.best_match(next_file, file_list)
                    paired_file = file_list.pop(match_ind)
                    file_queue.append((next_file, paired_file))
            else:
                for file_name in file_list:
                    file_queue.append((file_name, ))
        return file_queue
    
    @staticmethod
    def best_match(file_name: str, file_list: List[str]) -> int:
        def string_intersect(str_one, str_two):
            l = min(len(str_one), len(str_two))
            for i in range(l):
                if str_one[i] != str_two[i]:
                    return str_one[:i]
            else:
                return str_one[:l]

        match_ind = 0
        best_match = string_intersect(file_name, file_list[0])
        for i in range(1, len(file_list)):
            current_match = string_intersect(file_name, file_list[i])
            if len(current_match) > len(best_match):
                best_match = current_match
                match_ind = i
        return match_ind

    def combine_raw_read_output(self) -> None:
        temporary_files = ls(self.temp_dir, ext=".reads_%s" % EXPAM_TEMP_EXT)
        base_file_names = {
            re.match(r'(\S+)_\d+.reads_%s' % EXPAM_TEMP_EXT, os.path.basename(file_name)).group(1)
            for file_name in temporary_files
        }
        for sample_name in base_file_names:
            results_file_name = os.path.join(self.results_config.phy_raw, sample_name + ".csv")
            raw_files = [
                file for file in temporary_files
                if re.match(r'.*{base}_\d+.reads_{ext}$'.format(base=sample_name, ext=EXPAM_TEMP_EXT), file)
            ] 
            self.concatenate_files(results_file_name, *raw_files)
    
    @staticmethod
    def concatenate_files(dest, *files):
        with open(dest, 'w') as out_f:
            for file in files:
                # Write data to concatenated file.
                with open(file, 'r') as in_f:
                    out_f.write(in_f.read().strip() + "\n")
                # Remove temporary file.
                os.remove(file)

# base class =====================================================================

class _ClassificationOutput(metaclass=abc.ABCMeta):
    _c_perc = "Cumulative Classified Percentage"
    _c_cumul = "Cumulative Classified Count"
    _c_count = "Raw Classified Count"
    _s_perc = "Cumulative Split Percentage"
    _s_cumul = "Cumulative Split Count"
    _s_count = "Raw Split Count"
    _t_perc = "Cumulative Total Percentage"
    _t_cumul = "Cumulative Total Count"
    _t_count = "Raw Total Count"

    col_names = [_c_perc, _c_cumul, _c_count,
                 _s_perc, _s_cumul, _s_count,
                 _t_perc, _t_cumul, _t_count]
    
    def __init__(self, sl: pd.DataFrame, ml: pd.DataFrame):
        assert all(c1 == c2 for c1, c2 in zip(sl.columns, ml.columns))
        # `sl` and `ml` are both `pd.DataFrame` containing all samples as columns,
        # and every node as rows, including `unclassified`
        self.sl = sl
        self.ml = ml

    @abc.abstractmethod
    def write_sep(self, df: pd.DataFrame, out_file: str, sep: str = ",") -> None:
        """Write DataFrame to file.
        """
        df.to_csv(out_file, sep=sep, header=True)

    @staticmethod
    def remove_zeros(*dfs: pd.DataFrame):
        mask: pd.Series[bool] = None
        for df in dfs:
            non_zero = df.apply(lambda r: ((r > 0).any()) | (r.name == 'unclassified'), axis=1)
            mask = non_zero if mask is None else mask | non_zero
        for df in dfs:
            yield df.loc[mask, :]

    @abc.abstractmethod
    def accumulate(self, _df: pd.DataFrame) -> pd.DataFrame:
        """Accumulate counts within measurement units.
        """
        pass

    @staticmethod
    def get_cutoff_mask(_counts: pd.DataFrame, cpm: float) -> pd.DataFrame:
        min_counts = _counts.sum(axis=0) \
            .mul(cpm / 1e6) \
            .astype(_counts.values.dtype)
        # zero out counts below cutoff
        msk = _counts.lt(min_counts, axis=1)
        return msk
    
    @staticmethod
    def add_dfs(large: pd.DataFrame, small: pd.DataFrame) -> pd.DataFrame:
        assert all(small.columns.isin(large.columns))
        assert all(large.columns.isin(small.columns))
        assert all(small.index.isin(large.index))
        df = large.copy(deep=True)
        df.loc[small.index, small.columns] += small.values
        return df

    def summarise(self, output_base: str, cpm: float = 0.0, sep: str = ",") -> None:
        # accumulate counts
        sl_cumul = self.accumulate(self.sl)
        ml_cumul = self.accumulate(self.ml)
        # combine counts
        tot_count = self.add_dfs(self.sl, self.ml)
        tot_cumul = self.add_dfs(sl_cumul, ml_cumul)
        # apply cutoff
        min_counts = tot_count.sum(axis=0) \
            .mul(cpm / 1e6) \
            .astype(tot_cumul.values.dtype)
        msk = tot_cumul.lt(min_counts, axis=1)
        tot_cumul[msk] = 0
        # check structure of indices
        assert "unclassified" in self.sl.index
        assert self.sl.index.isin(sl_cumul.index).all()
        assert self.ml.index.isin(ml_cumul.index).all()
        assert ml_cumul.index.isin(sl_cumul.index).all()
        # write sample summary files
        for sample_name in self.sl.columns:
            df = self.summarise_sample(self.sl[sample_name], sl_cumul[sample_name],
                                       self.ml[sample_name], ml_cumul[sample_name],
                                       tot_count[sample_name], tot_cumul[sample_name],
                                       cpm=cpm)
            # save to disk
            sample_file = os.path.join(output_base, sample_name + ".csv")
            self.write_sep(df, sample_file, sep=sep)
            print("[formatter] wrote sample summary %s to %s" % (sample_name, sample_file))
        # write total cumulative to disk
        cumul_file = os.path.join(output_base, "cumulative.cpm_%.3f.csv" % cpm)
        self.write_sep(tot_cumul, cumul_file, sep=sep)
        print("[formatter] wrote cumulative counts with %.3f cutoff to %s" % (cpm, cumul_file))

    def summarise_sample(self,
                         sl: pd.Series, sl_cumul: pd.Series,
                         ml: pd.Series, ml_cumul: pd.Series,
                         tot: pd.Series, tot_cumul: pd.Series,
                         cpm: float = 0.0) -> pd.DataFrame:
        num_reads = sl.sum() + ml.sum()
        # prepare dataframe
        df = pd.DataFrame(0, index=sl_cumul.index, columns=self.col_names)
        # insert counts
        df.loc[sl.index, self._c_count] = sl
        df.loc[sl_cumul.index, self._c_cumul] = sl_cumul
        df.loc[ml.index, self._s_count] = ml
        df.loc[ml_cumul.index, self._s_cumul] = ml_cumul
        df.loc[tot.index, self._t_count] = tot
        df.loc[tot_cumul.index, self._t_cumul] = tot_cumul
        # employ cutoff
        df = df.loc[df[self._t_cumul] > 0, :]
        # format percentage signs
        df.loc[:, self._c_perc] = self.to_percentage(df.loc[:, self._c_cumul], num_reads)
        df.loc[:, self._s_perc] = self.to_percentage(df.loc[:, self._s_cumul], num_reads)
        df.loc[:, self._t_perc] = self.to_percentage(df.loc[:, self._t_cumul], num_reads)
        return df

    @staticmethod
    def to_percentage(_ser: pd.Series, total: int, decimals: int = 3) -> pd.Series:
        return _ser \
            .mul(100 / total) \
            .round(decimals) \
            .astype("string") \
            .add(" %")

# phylogenetic class =====================================================================

class PhylogeneticResults(_ClassificationOutput):
    def __init__(self, sl: pd.DataFrame, ml: pd.DataFrame, tree: Index):
        super(PhylogeneticResults, self).__init__(sl, ml)
        self.tree = tree

    @staticmethod
    def load_tree(tree: Index = None, phylogeny_path: str = None) -> Index:
        # load tree
        if tree is None and phylogeny_path is None:
            raise ValueError("Tree must be supplied!")
        elif tree is None:
            _, tree = Index.load_newick(phylogeny_path)
        return tree

    @staticmethod
    def make_index(tree: Index) -> List[str]:
        return ["unclassified" if i == 0 else node.name for i, node in enumerate(tree.pool)]
    
    @staticmethod
    def read_phy_output(csv_file: str, tree: Index, sep: str = ',') -> pd.DataFrame:
        ## read raw data
        #
        _data = pd.read_csv(csv_file, sep=sep, index_col=0, header=0)
        _data.index = [Index._format_node_name(v) for v in _data.index]
        ## target buffer
        #
        df = pd.DataFrame(0,
                          index=PhylogeneticResults.make_index(tree),
                          columns=_data.columns)
        df.loc[_data.index] = _data.values
        return df

    # read-wise parsing ---------------------------------------------------------------

    @classmethod
    def from_raw_read_output(cls, raw_read_base: str, num_procs: int, tree: Index = None, phylogeny_path: str = None):
        tree = PhylogeneticResults.load_tree(tree=tree, phylogeny_path=phylogeny_path)
        # read data
        output_files = ls(raw_read_base, ext=".csv")
        nodes = PhylogeneticResults.make_index(tree)
        ser_list = PhylogeneticResults.count_all_reads(output_files, nodes, num_procs)
        # sample-wise merge counts
        (_, sl_series_list, ml_series_list) = zip(*ser_list)
        sl = pd.concat(sl_series_list, axis=1)
        ml = pd.concat(ml_series_list, axis=1)
        return cls(sl, ml, tree)

    @staticmethod
    def count_all_reads(result_files: List[str], nodes: List[str], num_procs: int) -> List[Tuple[str, pd.Series, pd.Series]]:
        num_procs = min(num_procs, len(result_files))
        with Pool(processes=num_procs) as mp_pool:
            f = partial(PhylogeneticResults.count_sample_reads,
                        nodes=nodes)
            ser_list = mp_pool.map_async(f, result_files).get()
        return sorted(ser_list, key = lambda t: t[0])
    
    @staticmethod
    def count_sample_reads(file: str, nodes: List[str]) -> Tuple[str, pd.Series, pd.Series]:
        name = os.path.basename(file)[:-4]
        sl = pd.Series(0, index=nodes, name=name)
        ml = pd.Series(0, index=nodes, name=name)
        with open(file, 'r') as f:
            for line in f:
                data = line.split("\t")
                pid = int(data[2][1:])
                if data[0] == 'S':
                    ml.iloc[pid] += 1
                else:
                    sl.iloc[pid] += 1
        return name, sl, ml

    # preloaded -------------------------------------------------------------------------

    @classmethod
    def from_tables(cls, sl_file: str, ml_file: str, sep: str = ',',
                    tree: Index = None, phylogeny_path: str = None):
        ## load tree
        #
        tree = PhylogeneticResults.load_tree(tree=tree, phylogeny_path=phylogeny_path)
        ## data parsing
        #
        sl = PhylogeneticResults.read_phy_output(sl_file, tree, sep=sep)
        ml = PhylogeneticResults.read_phy_output(ml_file, tree, sep=sep)
        return cls(sl, ml, tree)

    # virtual methods ---------------------------------------------------------------

    def accumulate(self, _df: pd.DataFrame) -> pd.DataFrame:
        df = _df.copy()
        for node in self.tree.pool[1:][::-1]:
            df.loc[node.name, :] += df.loc[node.children, :].sum(axis=0)
        return df

    def write_sep(self, df: pd.DataFrame, out_file: str, sep: str = ","):
        df = df.copy()
        df.index = [
            'p' + pid
            if (pid != "unclassified") and (self.tree[pid].type == "Branch")
            else pid
            for pid in df.index
        ]
        _ClassificationOutput.write_sep(self, df, out_file, sep=sep)

    # helpers ---------------------------------------------------------------------------

    def to_csv(self, sl_file: str, ml_file: str, keep_zeros: bool = False):
        sl = self.sl.copy()
        ml = self.ml.copy()
        # remove null rows
        if not keep_zeros:
            (sl, ml) = self.remove_zeros(sl, ml)
        if "unclassified" in ml.index:
            ml.drop(index="unclassified", inplace=True)
        # save counts to disk
        self.write_sep(sl, sl_file)
        self.write_sep(ml, ml_file)

    # plotting ---------------------------------------------------------------------------

    def draw_results(self,
                     output_base: str,
                     groups: List[Tuple[str, Tuple[str, ...]]] = None,
                     cpm: float = 0.0,
                     colour_list: List[str] = None, keep_zeros: bool = True,
                     use_node_names: bool = True, log_scores: bool = True,
                     itol_mode: bool = False, flat_colour: bool = False):
        ## prepare dataframes
        #
        #   - SL
        sl_cumul = self.accumulate(self.sl)
        #   - ML
        ml_cumul = self.accumulate(self.ml)
        #   - total
        tot_cumul = sl_cumul.copy()
        tot_cumul.loc[ml_cumul.index, :] += ml_cumul.values

        ## render
        #
        # classified results
        phy_pdf = os.path.join(output_base, "phylotree_classified.pdf")
        self.draw_counts(sl_cumul, phy_pdf, inplace=True, groups=groups,
                         cpm=cpm, colour_list=colour_list,
                         keep_zeros=keep_zeros, use_node_names=use_node_names,
                         log_scores=log_scores, itol_mode=itol_mode, flat_colour=flat_colour)
        # split results
        split_pdf = os.path.join(output_base, "phylotree_splits.pdf")
        self.draw_counts(ml_cumul, split_pdf, inplace=True, groups=groups,
                         cpm=cpm, colour_list=colour_list,
                         keep_zeros=keep_zeros, use_node_names=use_node_names,
                         log_scores=log_scores, itol_mode=itol_mode, flat_colour=flat_colour)
        # total counts
        total_pdf = os.path.join(output_base, "phylotree_total.pdf")
        self.draw_counts(tot_cumul, total_pdf, inplace=True, groups=groups,
                         cpm=cpm, colour_list=colour_list,
                         keep_zeros=keep_zeros, use_node_names=use_node_names,
                         log_scores=log_scores, itol_mode=itol_mode, flat_colour=flat_colour)

    def draw_counts(self,
                    counts: pd.DataFrame, out_file: str, inplace: bool = True,
                    groups: List[Tuple[str, Tuple[str, ...]]] = None,
                    cpm: float = 0.0,
                    colour_list: List[str] = None, keep_zeros: bool = True,
                    use_node_names: bool = True, log_scores: bool = True,
                    itol_mode: bool = False, flat_colour: bool = False):
        if not inplace:
            counts = counts.copy()
        ## cutoff
        #
        min_count_per_sample = counts.sum(axis=0) \
            .mul(cpm / 1e6) \
            .astype(counts.values.dtype)

        # zero out counts below cutoff
        msk = counts.lt(min_count_per_sample, axis=1)
        counts[msk] = 0

        ## render
        #
        clades_with_counts = (counts > 0).any(axis=1).sum()
        if clades_with_counts > 0:
            self.tree.draw_counts(counts, out_file, groups=groups, colour_list=colour_list,
                                  keep_zeros=keep_zeros, use_node_names=use_node_names,
                                  log_scores=log_scores, itol_mode=itol_mode,
                                  flat_colour=flat_colour)
        else:
            print("Skipping plotting of %s - no samples with counts in this matrix." % out_file)

# taxonomic class =====================================================================

class TaxonomicResults(_ClassificationOutput):
    _rank = "Rank"
    _scientific_name = "Scientific Name"

    def __init__(self, sl: pd.DataFrame, ml: pd.DataFrame, taxonomy: TaxonomyInterface):
        super(TaxonomicResults, self).__init__(sl, ml)
        self.taxonomy = taxonomy
        required_taxids = {
            ptid
            for taxid in self.ml.index
            if taxid != "unclassified"
            for ptid in self.taxonomy.get_lineage(taxid)
        }
        self.taxid2rank = self.taxonomy.get_ranks(*required_taxids)
        self.taxid2name = self.taxonomy.get_names(*required_taxids)

    # hand-over -------------------------------------------------------------------------

    @classmethod
    def from_phylogenetic(cls, _data: PhylogeneticResults, taxonomy: TaxonomyInterface) :
        # convert to taxonomic counts
        t_cx_df = TaxonomicResults.gather_by_taxid(_data.sl.copy(deep=True), taxonomy)
        t_spx_df = TaxonomicResults.gather_by_taxid(_data.ml.copy(deep=True), taxonomy)
        return cls(t_cx_df, t_spx_df, taxonomy)

    # preloaded -------------------------------------------------------------------------

    @classmethod
    def from_tables(cls, sl_file: str, ml_file: str, tree: Index, taxonomy: TaxonomyInterface, sep: str = ','):
        # load phylogenetic output
        cx_df = PhylogeneticResults.read_phy_output(sl_file, tree, sep=sep)
        spx_df = PhylogeneticResults.read_phy_output(ml_file, tree, sep=sep)
        # convert to taxonomic counts
        t_cx_df = TaxonomicResults.gather_by_taxid(cx_df, taxonomy)
        t_spx_df = TaxonomicResults.gather_by_taxid(spx_df, taxonomy)
        return cls(t_cx_df, t_spx_df, taxonomy)

    @staticmethod
    def gather_by_taxid(df: pd.DataFrame, taxonomy: TaxonomyInterface) -> pd.DataFrame:
        # combine counts within taxid
        df["taxid"] = df.index \
            .to_series() \
            .map(taxonomy.clade2taxid)
        df = df \
            .groupby(["taxid"], axis=0) \
            .sum()
        assert df.index.name == "taxid"
        # sort index
        _order = {taxid: i+1 for i, taxid in enumerate(taxonomy.taxid_pool)}
        _order["unclassified"] = 0
        df.sort_index(inplace=True, key=lambda _i: _i.map(_order))
        return df
    
    # virtual methods ------------------------------------------------------------------

    def accumulate(self, _df: pd.DataFrame) -> pd.DataFrame:
        ## ensure lineage completeness
        #
        current_taxids = set(_df.index.values)
        current_taxids.remove("unclassified")
        required_taxids = {
            ptid
            for taxid in current_taxids
            for ptid in self.taxonomy.get_lineage(taxid)
        }
        # add root
        if "1" not in current_taxids:
            required_taxids.add("1")
        # append missing taxids
        missing_taxids = list(required_taxids - current_taxids)
        sec = pd.DataFrame(0, index=missing_taxids, columns=_df.columns)
        df = pd.concat([_df, sec])
        ## accumulate counts
        #
        for tid in self.taxonomy.taxid_pool[::-1]:
            if tid in df.index:
                for cid in self.taxonomy.taxid2child_taxids[tid]:
                    if cid in df.index:
                        df.loc[tid, :] += df.loc[cid, :]
        ## sort result
        #
        _order = {taxid: i+1 for i, taxid in enumerate(self.taxonomy.taxid_pool)}
        _order["unclassified"] = 0
        df.sort_index(inplace=True, key=lambda _i: _i.map(_order))
        return df
    
    def attach_taxonomic_data(self, df: pd.DataFrame) -> pd.DataFrame:
        has_taxonomy = df.index.isin(self.taxid2rank.keys()) & df.index.isin(self.taxid2name.keys())
        df.loc[has_taxonomy, self._rank] = df.index[has_taxonomy] \
            .to_series() \
            .map(self.taxid2rank)
        df.loc[has_taxonomy, self._scientific_name] = df.index[has_taxonomy] \
            .to_series() \
            .map(self.taxid2name)
        return df

    def write_sep(self, df: pd.DataFrame, out_file: str, sep: str = ","):
        df = self.attach_taxonomic_data(df)
        _ClassificationOutput.write_sep(self, df, out_file, sep=sep)

    def convert_raw(self, input_base: str, output_base: str, tree: Index):
        for read_file in ls(input_base, ext=".csv"):
            out_file = os.path.join(output_base, os.path.basename(read_file))
            with open(read_file, 'r') as infh, open(out_file, 'w') as outfh:
                for line in infh:
                    read_data = line \
                        .strip() \
                        .split("\t")[:-1]
                    if read_data[0] == 'U':
                        read_data[2] = "NA"
                    else:
                        pid = int(read_data[2][1:])
                        clade_name = tree.pool[pid].name
                        read_data[2] = self.taxonomy.clade2taxid[clade_name]
                    outfh.write("\t".join(read_data) + "\n")
            print("[taxonomy] mapped read-wise output from %s to %s" % (read_file, out_file))



# version: 2
import logging
import os
from typing import List, Set, Tuple
from expam.classify.classify import TaxonomicResults, run_classifier
from expam.classify.config import create_tax_results, make_results_config, validate_results_configuration
from expam.classify.taxonomy import TaxonomyInterface
from expam.cli.main import CommandGroup, ExpamOptions, clear_logs
from expam.database import FileLocationConfig
from expam.database.config import JSONConfig, make_database_config, validate_database_file_configuration
from expam.tree.tree import Index
from expam.utils import die, is_hex
import expam.logger

class ClassifyCommand(CommandGroup):
    commands: Set[str] = {
        'classify', 'to_taxonomy'
    }

    def __init__(
        self, config: FileLocationConfig,
        files: List[str], out_dir: str,
        convert_to_taxonomy: bool, cpm: float, groups: List[Tuple[str]],
        use_node_names: bool, keep_zeros: bool,
        colour_list: List[str], paired_end: bool, alpha: float,
        log_scores: bool, itol_mode: bool, flat_colour: bool, debug: bool
    ) -> None:
        super().__init__()

        self.config = config
        self.results_config = make_results_config(out_dir)

        self.files = files
        self.out_dir = out_dir

        self.convert_to_taxonomy = convert_to_taxonomy
        self.cpm = cpm
        self.groups = groups

        self.use_node_names = use_node_names
        self.keep_zeros = keep_zeros
        self.colour_list = colour_list

        self.paired_end = paired_end
        self.alpha = alpha
        self.log_scores = log_scores
        self.itol_mode = itol_mode
        self.flat_colour = flat_colour

        if debug:
            expam.logger.current_logging_level = logging.DEBUG

    @classmethod
    def take_args(cls: CommandGroup, args: ExpamOptions) -> dict:
        cpm, alpha = cls.parse_floats(args.cpm, args.alpha)

        if args.out_url is None:
            die("Must supply -o/--out.")

        # Format groups.
        if args.groups is not None:
            groups = [v for v in args.groups if v]

            for i in range(len(groups)):
                if is_hex(groups[i][0]):
                    groups[i] = (groups[i][0], tuple(groups[i][1:]))
                else:
                    groups[i] = (None, tuple(groups[i]))
        else:
            groups = None

        # Check colour list.
        if args.colour_list is not None:
            colour_list = [hex for hex in args.colour_list if is_hex(hex)]

            if not colour_list:
                colour_list = None
        else:
            colour_list = None

        return {
            'config': make_database_config(args.db_name),
            'files': args.directory,
            'out_dir': args.out_url,
            'convert_to_taxonomy': args.taxonomy,
            'cpm': cpm,
            'groups': groups,
            'use_node_names': not args.ignore_names,
            'keep_zeros': args.keep_zeros,
            'colour_list': colour_list,
            'paired_end': args.paired_end,
            'alpha': alpha,
            'log_scores': args.log_scores,
            'itol_mode': args.itol_mode,
            'flat_colour': args.flat_colour,
            'debug': args.debug
        }

    def check_database_exists(self):
        validate_database_file_configuration(self.config)
        if not os.path.exists(self.config.database_file):
            die("Database has not been built! Not found at %s." % self.config.database_file)

    """
    Run command
    ===========
    
    """
    def classify(self):
        self.check_database_exists()
        clear_logs(self.config.logs)

        # Read configuration file.
        config = JSONConfig(self.config.conf)
        k, n, phylogeny_path, _, _ = config.get_build_params()
        keys_shape, values_shape = tuple(config["keys_shape"]), tuple(config["values_shape"])

        # Correct for trimmed keys_shape.
        if len(keys_shape) == 1:
            keys_shape = keys_shape + (1,)

        # Run expam classification.
        run_classifier(self.files, self.out_dir, self.config.base, k, n-1, phylogeny_path,
                       keys_shape, values_shape, self.config.logs,
                       taxonomy=self.convert_to_taxonomy, groups=self.groups,
                       keep_zeros=self.keep_zeros, cpm=self.cpm, use_node_names=self.use_node_names,
                       colour_list=self.colour_list, paired_end=self.paired_end, alpha=self.alpha,
                       log_scores=self.log_scores, itol_mode=self.itol_mode, flat_colour=self.flat_colour)
    
    """
    Convert phylogenetic results to taxonomy
    ========================================
    
    """
    def to_taxonomy(self):
        self.check_database_exists()
        if self.out_dir is None:
            die("Require output directory (-o, --out_dir)!")
        else:
            validate_results_configuration(self.results_config)
        # load phylogenetic tree
        config = JSONConfig(self.config.conf)
        phylogeny_path = config.get_phylogeny_path()
        _, tree = Index.load_newick(phylogeny_path)
        # create phylogenetic results
        create_tax_results(self.results_config)
        my_ncbi = TaxonomyInterface(self.config, tree=tree)
        tax = TaxonomicResults.from_tables(self.results_config.phy_classified,
                                           self.results_config.phy_split,
                                           tree,
                                           my_ncbi)
        tax.summarise(self.results_config.tax,
                      cpm=self.cpm)
        tax.convert_raw(self.results_config.phy_raw, self.results_config.tax_raw, tree)



import gzip
import io
import logging
import os
import random
import re
from tempfile import NamedTemporaryFile
import warnings
import tables as tb
from expam.ext.kmers import reverse_complement_combine
from expam.logger import timeit

warnings.filterwarnings('ignore', category=tb.NaturalNameWarning)

COMPRESSION_EXTNS = ['.tar.gz', '.tar', '.gz']
DEFAULT_MODE = "rb"
DEFAULT_OPENER = open
COMP_PARSE = {
    ".tar.gz": {"mode": "rb", "opener": gzip.open},
    ".gz": {"mode": "rb", "opener": gzip.open}
}

def get_opener(file_name):
    for suffix in COMP_PARSE:

        if check_suffix(file_name, suffix):
            file_name = file_name.replace(suffix, '')
            mode, opener = COMP_PARSE[suffix]["mode"], COMP_PARSE[suffix]["opener"]

            break

    else:
        mode, opener = DEFAULT_MODE, DEFAULT_OPENER

    return file_name, (mode, opener)


def check_suffix(string: str, sfx: str):
    string = string.strip()

    if string.endswith(sfx):
        return True
    return False


def format_tree_string(nwk: str) -> str:
    seq_types = [".fna", ".faa"]
    comp_types = [".tar.gz", ".gz", ""]
    nwk = nwk.replace("'", "")
    for seq_type in seq_types:
        for comp_type in comp_types:
            ext = seq_type + comp_type + ":"
            if ext in nwk:
                nwk = nwk.replace(ext, ":")
    return nwk


def format_name(name, remove_comp=False):
    name = os.path.basename(name)
    # Must not be a digit. If it is, add 'ref' to the start.
    if name.isdigit():
        name = "ref" + name

    # Extract sequence name.
    if ".fna" in name:
        name = name.replace(".fna", "")
    elif ".faa" in name:
        name = name.replace(".faa", "")

    if remove_comp:
        for suffix in COMP_PARSE:
            if check_suffix(name, suffix):
                name = name.replace(suffix, '')

                break

    return name


def extract_id(metadata):
    # Check for kraken format.
    if "kraken" in metadata:
        start = metadata.rfind("|") + 1
    else:
        # Standard fasta format.
        if ">" in metadata:
            start = metadata.find(">") + 1
        elif "@" in metadata:
            start = metadata.find("@") + 1

    end = metadata.find(" ")
    if not end:  # Case where no metadata is provided.
        end = len(metadata)

    # Contig name, including potential contig version number.
    contig_id = metadata[start:end]
    # Concatenate multiple contig versions.
    version_start = contig_id.rfind(".")
    if version_start:
        contig_id = contig_id[:version_start]
    return contig_id


@timeit
def iterate_sequence(url):
    with open(url, "r") as f:
        seqID, contig = "", ""
        for sequence in f:
            # Strip newlines.
            if '\n' in sequence:
                sequence = sequence.strip('\n')
            if ">" in sequence:
                if contig != "":
                    # Return the previous sequence, if there is one.
                    yield (seqID, contig)
                # Find seqID.
                if "|" in sequence:
                    # >x|seqID|... format.
                    seqID = sequence.split("|")[1]
                else:
                    # >seqID ... format.
                    seqID = sequence.split(" ")[0]
            else:
                # Combine sequence.
                contig += sequence
        # End case: yeild last contig.
        if contig != "":
            yield seqID, contig


@timeit
def get_sequence(url):
    """
    Extract the sequence from FASTA source file @ url.

    :param url: String.
    :return: String.
    """
    sequence = io.BytesIO()
    _, (mode, opener) = get_opener(url)

    with opener(url, mode) as f:
        for line in f:
            line = line.strip()
            # Skip information line at start of FASTA format.
            if b">" in line:
                sequence.write(b"N")

            else:
                sequence.write(line)

    return sequence.getvalue()

class MaskStore:
    def __init__(self):
        self.store = {}

    def __setitem__(self, key, value):
        # Place value in store.
        self.store[key] = value

    def __getitem__(self, key):
        return self.store[key]

    def __delitem__(self, key):
        del self.store[key]

class DiskMaskStore:
    def __init__(self):
        self._file = NamedTemporaryFile(mode="w", suffix=".h5")
        self._dbh = tb.open_file(self._file.name, mode="w")

    def __setitem__(self, key, value):
        # string key maps to UInt32 array
        self._dbh.create_array(self._dbh.root, key, value)

    def __getitem__(self, key):
        return self._dbh.root[key].read()
    
    def __delitem__(self, *args, **kwargs):
        # PyTables does not support on-disk array deletion
        pass

    def __del__(self):
        self._dbh.close()
        self._file.close()

class _SequenceStore:
    def __init__(self, k: int, out_dir: str, logger: logging.Logger) -> None:
        self.stream = io.BytesIO()
        self.k = k
        self.out_dir = out_dir
        self.accession_ids = {}
        self.logger = logger

    @staticmethod
    def extract_accession_id(string):
        def try_convert(byte_string):
            try:
                return byte_string.decode("utf8")

            except AttributeError:
                return byte_string

        accn_id = re.findall(b'(?:^\s*\>)(.*?)(?:\s+|\|)', string)
        tax_id = re.findall(b'(?:\|kraken:taxid\||\|taxid\|)(\d+)(?:\s+|\|)', string)

        accn_id = None if not accn_id else try_convert(accn_id[0])
        tax_id = None if not tax_id else try_convert(tax_id[0])

        return accn_id, tax_id

    def _read_sequence(self, file_path: str):
        # Check it is a valid file.
        if not os.path.isfile(file_path):
            raise ValueError("Directory %s is not a file!" % file_path)
        
        self.logger.info("extracting sequences from %s" % file_path)

        # Determine opening mode based on compression of sequence file.
        file_name = format_name(file_path)
        file_name, (mode, opener) = get_opener(file_name)

        with opener(file_path, mode=mode) as f:
            # Extract the accession id.
            metadata = f.readline()
            assert b'>' in metadata

            accession_id, tax_id = self.extract_accession_id(metadata)

            if accession_id is None and tax_id is None:
                raise ValueError("Couldn't identify sequence %s!" % file_name)

            for line in f:
                line = line.strip()
                # Skip information line at start of FASTA format.
                if b'>' in line:
                    self.stream.write(b'N')
                else:
                    self.stream.write(line)

        # Convert to numeric sequence and put in store.
        sequence_view = self.stream.getvalue()

        # Save accession ID for this file.
        self.accession_ids[file_name] = [accession_id, tax_id]

        # Clear stream.
        self.stream.seek(0)
        self.stream.truncate()

        return file_name, sequence_view

    def save_accession_ids(self, suffix=""):
        id_data = "\n".join([
            "%s,%s,%s" % (name, keys[0], keys[1])
            for name, keys in self.accession_ids.items()
        ])

        with open(os.path.join(self.out_dir, "phylogeny", "accession_ids_%s.csv" % suffix), "w") as f:
            f.write(id_data)

class SequenceStore(_SequenceStore):
    def __init__(self, k: int, out_dir: str, logger: logging.Logger):
        super().__init__(k, out_dir, logger=logger)
        self._store = {}

    def add_sequence(self, file_path):
        file_name, sequence = self._read_sequence(file_path)
        self._store[file_name] = sequence
        return file_name, len(sequence)

    def __getitem__(self, key):
        # If it is a branch, this is a valid request,
        # but no information from the Sequence Store is required.
        if key.isdigit():
            return None
        else:
            return self._store.__getitem__(key)

    def __delitem__(self, key):
        self._store.__delitem__(key)

    def keys(self):
        return self._store.keys()

class DiskSequenceStore(_SequenceStore):
    def __init__(self, k: int, out_dir: str, logger: logging.Logger):
        super().__init__(k, out_dir, logger=logger)
        self._file = NamedTemporaryFile(mode="w", suffix=".h5")
        self._dbh = tb.open_file(self._file.name, mode="w")

    def add_sequence(self, file_path):
        file_name, sequence = self._read_sequence(file_path)
        self._dbh.create_array(self._dbh.root, file_name, sequence)
        return file_name, len(sequence)
    
    def __getitem__(self, key: str):
        # If it is a branch, this is a valid request,
        # but no information from the Sequence Store is required.
        if key.isdigit():
            return None
        else:
            return self._dbh.root[key].read()

    def __delitem__(self, *args, **kwargs):
        pass

    def keys(self):
        return self._dbh.root._v_leaves.keys()

    def __del__(self):
        self._dbh.close()
        self._file.close()


def make_reads(in_url, l, n, out_url, e=0.0):
    """
    Make n reads from sequence at url, from a expam maker class.
    """
    maker = Reads(l, out_url)
    maker.make(in_url, n, error=e)


class Reads:
    def __init__(self, l, out_dir):
        # Length of reads.
        self.l = l
        # Directory to save reads.
        self.out_dir = out_dir

    def make(self, f_url, n_reads, error=0.0):
        """
        Produce a number of FASTQ reads from a sequence.

        :param error: Float - Rate that errors will be put into reads (error ~ reads with errors / reads).
        :param f_url: String - Url of the sequence.
        :param n_reads: Int - number of reads to produce.
        """
        if error > 1:
            if error > 100:
                raise ValueError("Error rate must be either < 1 or < 100!")

            error /= 100

        nt_codes = {0: b"A", 1: b"C", 2: b"G", 3: b"T"}
        bad_quality = bytes("N" * 75, encoding="utf8")

        # Check formatting of sequence.
        self.check_fna(f_url)
        # Extract sequence string from f_url.
        sequence = get_sequence(f_url)

        # Extract the file name from the path.
        forward_name, reverse_name = self.get_file_name(f_url)

        # Generate the reads.
        i = 0
        forward_read_stream = io.BytesIO()
        reverse_read_stream = io.BytesIO()

        while i < n_reads:

            # Generate a start index for the read.
            index = random.randint(0, len(sequence) - self.l)
            # Get read.
            read = sequence[index: index + self.l]

            # Check quality.
            if bad_quality in read:
                # This is deemed poor quality. Try again.
                continue

            rid = ("R" + str(abs(hash(str(index))))).encode('utf8')
            meta_line = b">%b Read from %b.\n" % (rid, f_url.encode('utf8'))

            # Generate error.
            has_error = random.randint(0, 100) / 100 < error

            if has_error:
                sub_location = random.randint(0, len(read) - 1)
                nt = nt_codes[random.randint(0, 3)]

                read_string = b"%b%b%b" % (read[:sub_location], nt, read[sub_location + 1:])

            else:
                read_string = b"%b" % read

            forward_reverse_string = reverse_complement_combine(read_string, read_string)
            l = len(read_string)

            forward_string = forward_reverse_string[:l]
            reverse_string = forward_reverse_string[-l:]

            forward_read_stream.write(meta_line)
            forward_read_stream.write(forward_string)
            forward_read_stream.write(b'\n')

            reverse_read_stream.write(meta_line)
            reverse_read_stream.write(reverse_string)
            reverse_read_stream.write(b'\n')

            i += 1

        # Write to file.
        with open(os.path.join(self.out_dir, forward_name), "wb") as f:
            f.write(forward_read_stream.getvalue())

        forward_read_stream.seek(0)
        forward_read_stream.truncate()

        with open(os.path.join(self.out_dir, reverse_name), "wb") as f:
            f.write(reverse_read_stream.getvalue())

        reverse_read_stream.seek(0)
        reverse_read_stream.truncate()

    @staticmethod
    def check_fna(url):
        if not ".fna" in url:
            raise TypeError("Sequence type must be .fna.")

    @staticmethod
    def get_file_name(url):
        path = os.path.normpath(url)
        path = path.split(os.sep)

        forward_name = path[-1].rstrip(".fna") + "_1.fa"
        reverse_name = path[-1].rstrip(".fna") + "_2.fa"

        return forward_name, reverse_name



# version: 2
from __future__ import annotations
import os
from typing import List, Mapping, Set, Tuple, Type, TypeVar
from ete3 import NCBITaxa

from expam.database import FileLocationConfig
from expam.database.config import JSONConfig
from expam.tree.tree import Index
from expam.utils import work_in_directory

# NCBI -----------------------------------------------------------------------

class ncbi_taxdump:
    _remote_dump_url = "http://ftp.ncbi.nih.gov/pub/taxonomy/taxdump.tar.gz"
    _sqlite_file_name = "ncbi.sqlite"
    _taxdump_file_name = "ncbi.dmp.tar.gz"

    # initialisation --------------------------------------------------------

    def __init__(self, taxonomy_dir: str, force_update: bool = False) -> None:
        self.taxdump = os.path.abspath(os.path.join(taxonomy_dir, self._taxdump_file_name))
        self.sqlite = os.path.abspath(os.path.join(taxonomy_dir, self._sqlite_file_name))
        self._check_sql_db(force_update)

    def log(self, msg: str) -> None:
        print("[taxdump] %s" % msg)
        return

    @staticmethod
    def contains_taxonomic_database(taxonomy_dir: str) -> bool:
        sqlite_file = os.path.abspath(os.path.join(taxonomy_dir,
                                                   ncbi_taxdump._sqlite_file_name))
        return os.path.exists(sqlite_file)

    def _check_sql_db(self, force_update: bool) -> None:
        if not os.path.exists(self.sqlite) or force_update:
            from ete3.ncbi_taxonomy.ncbiquery import update_db
            with work_in_directory(os.path.dirname(self.taxdump)):
                self._download_taxonomy_dump()
                update_db(self.sqlite, targz_file=self.taxdump)
                os.remove(self.taxdump)
        return

    def _download_taxonomy_dump(self) -> None:
        try:
            from urllib import urlretrieve
        except ImportError:
            from urllib.request import urlretrieve
        self.log('downloading taxdump.tar.gz from NCBI FTP site (via HTTP) to %s' % self.taxdump)
        urlretrieve(self._remote_dump_url, self.taxdump)
        self.log('done. Parsing...')
        return

    # load ----------------------------------------------------------------

    def load(self) -> NCBITaxa:
        return NCBITaxa(self.sqlite)

class ncbi:
    def __init__(self,
                 taxonomy_dir: str,
                 force_update: bool = False):
        # parse taxdump
        dmp = ncbi_taxdump(taxonomy_dir, force_update=force_update)
        self._sql_handle = dmp.load()

    def get_lineage(self, taxid: str | int) -> List[str]:
        return [str(v) for v in self._sql_handle.get_lineage(taxid)]

    def get_lineages(self, *taxids: str | int) -> Mapping[str, List[str]]:
        return {taxid: self.get_lineage(taxid) for taxid in taxids}

    def get_ranks(self, *taxids: str) -> Mapping[str, str]:
        return self._sql_handle.get_rank(taxids)

    def get_names(self, *taxids: str) -> Mapping[str, str]:
        return self._sql_handle.get_taxid_translator(taxids)

# mapping ------------------------------------------------------------------------

T = TypeVar('T')

class NodeMapper:
    def __init__(self,
                 name2taxid: Mapping[str, str],
                 taxid2lineage: Mapping[str, List[str]],
                 tree: Index = None,
                 db: FileLocationConfig = None,
                 phylogeny_path: str = None) -> None:
        self.name2taxid = name2taxid
        self.taxid2lineage = taxid2lineage
        self.tree = NodeMapper.load_tree(db=db, tree=tree, phylogeny_path=phylogeny_path)
        # auxilliary data strutures
        self.taxid_pool: List[str] = []
        self.taxid2child_taxids: Mapping[str, Set[str]] = {}

    @classmethod
    def from_ncbi(cls: Type[NodeMapper],
                  ncbi_handle: ncbi,
                  tree: Index = None,
                  db: FileLocationConfig = None,
                  phylogeny_path: str = None)-> NodeMapper:
        name2taxid = NodeMapper.load_genome_taxids(db.accn_id)
        taxid2lineage = ncbi_handle.get_lineages(*set(name2taxid.values()))
        return cls(name2taxid, taxid2lineage, tree=tree, db=db, phylogeny_path=phylogeny_path)

    @staticmethod
    def load_tree(db: FileLocationConfig = None, tree: Index = None, phylogeny_path: str = None):
        # load tree
        if db is None and tree is None and phylogeny_path is None:
            raise ValueError("Must supply a tree!")
        elif tree is None:
            if phylogeny_path is None:
                conf = JSONConfig(db.conf)
                phylogeny_path = conf.get_phylogeny_path()
            _, tree = Index.load_newick(phylogeny_path)
        return tree

    @staticmethod
    def load_genome_taxids(accession_ids_file: str) -> Mapping[str, str]:
        name2taxid = {}
        missing_taxids = 0
        with open(accession_ids_file, 'r') as f:
            for line in f:
                try:
                    genome, _, taxid = line.strip() \
                        .split(",")
                except KeyError:
                    continue
                if (len(taxid) == 0) or (taxid == "None"):
                    missing_taxids += 1
                else:
                    name2taxid[genome] = taxid
        if missing_taxids > 0:
            raise ValueError("%d genomes missing taxonomic assignments. \
                             Update %s accordingly." % (missing_taxids, accession_ids_file))
        return name2taxid

    def _init_taxid_aux(self):
        self.taxid_pool = []
        self.taxid2child_taxids = {}

    def map_all_nodes(self) -> Tuple[Mapping[str, str], List[str], Mapping[str, Set[str]]]:
        nodes = [node.name for node in self.tree.pool[1:]]
        return self.map_nodes(nodes)

    def map_nodes(self, phy_nodes) -> Tuple[Mapping[str, str], List[str], Mapping[str, Set[str]]]:
        # Map phylogeny index to taxonomy.
        clade2taxid: Mapping[str, str] = {}
        for clade in phy_nodes:
            taxid = self.map_phylogeny_node(clade)
            clade2taxid[clade] = taxid
        # shift container owners
        taxid_pool__export = self.taxid_pool
        taxid2child_taxids__export = self.taxid2child_taxids
        self._init_taxid_aux()
        return clade2taxid, taxid_pool__export, taxid2child_taxids__export
    
    def map_phylogeny_node(self, node_name: str) -> str:
        # get leaves within this clade and their taxids
        taxids = [self.name2taxid[leaf] for leaf in self.tree.yield_leaves(node_name)]
        lineages = [self.taxid2lineage[taxid] for taxid in taxids]
        # find intersection of lineages of these taxids
        common_lineage = self.tuple_intersect(*lineages)
        assert len(common_lineage) >= 1
        taxid = common_lineage[-1]
        # insert lineage into taxid pool
        pool_index = 0
        for i, parent_taxid in enumerate(common_lineage):
            # insert into taxid pool
            try:
                pool_index = self.taxid_pool.index(parent_taxid) + 1
            except ValueError:
                self.taxid_pool.insert(pool_index, parent_taxid)
                pool_index += 1
            # track immediate child nodes
            if i + 1 < len(common_lineage):
                child_taxid = common_lineage[i + 1]
                try:
                    self.taxid2child_taxids[parent_taxid].add(child_taxid)
                except KeyError:
                    self.taxid2child_taxids[parent_taxid] = {child_taxid}
            elif parent_taxid not in self.taxid2child_taxids:
                self.taxid2child_taxids[parent_taxid] = set()
        return taxid
    
    @staticmethod
    def tuple_intersect(*tuples: List[T]) -> Tuple[T, ...]:
        intersection = ()
        length = min(len(tup) for tup in tuples)
        for i in range(length):
            item = tuples[0][i]
            for j in range(1, len(tuples)):
                if tuples[j][i] != item:
                    return intersection
            else:
                intersection += (item,)
        return intersection

# Expam -----------------------------------------------------------------------

class TaxonomyInterface:
    def __init__(self,
                 conf: FileLocationConfig,
                 tree: Index = None,
                 phylogeny_path: str = None,
                 force_download: bool = False,
                 force_remap: bool = False):
        self._ncbi = ncbi(conf.phylogeny, force_update=force_download)
        if force_download \
            or force_remap \
            or any(not os.path.exists(getattr(conf, attr)) \
                   for attr in ('clade_tax', 'tax_pool', 'tax_childs')):
            # create and save map files
            op = NodeMapper.from_ncbi(self._ncbi, tree=tree, db=conf, phylogeny_path=phylogeny_path)
            self.clade2taxid, self.taxid_pool, self.taxid2child_taxids = op.map_all_nodes()
            self.clade2taxid["unclassified"] = "unclassified"
            self.export(conf, op.tree)
        else:
            # load map files
            self.clade2taxid = self.load__clade2taxid(conf)
            self.taxid_pool = self.load__taxid_pool(conf)
            self.taxid2child_taxids = self.load__taxid2child_taxids(conf)

    def export(self, conf: FileLocationConfig, tree: Index) -> None:
        # CLADE_TAX
        with open(conf.clade_tax, "w") as f:
            for clade, taxid in self.clade2taxid.items():
                f.write("%s,%s\n" % (clade, taxid))
        # TAX_POOL
        with open(conf.tax_pool, "w") as f:
            f.write("\n".join(self.taxid_pool))
        # TAX_CHILDS
        with open(conf.tax_childs, "w") as f:
            for taxid, children in self.taxid2child_taxids.items():
                f.write("%s,%s\n" % (taxid, ";".join(children)))
        # CLADE_TABLE
        self.write_clade_table(conf.clade_table, tree)

    @staticmethod
    def load__clade2taxid(conf: FileLocationConfig) -> Mapping[str, str]:
        clade2tax = {}
        with open(conf.clade_tax, "r") as f:
            for line in f:
                try:
                    clade, taxid = line.strip()\
                        .split(',')
                except ValueError:
                    continue
                clade2tax[clade] = taxid
        return clade2tax
    
    @staticmethod
    def load__taxid_pool(conf: FileLocationConfig) -> List[str]:
        with open(conf.tax_pool, "r") as f:
            return [line.strip() for line in f]
        
    @staticmethod
    def load__taxid2child_taxids(conf: FileLocationConfig) -> Mapping[str, Set[str]]:
        taxid2child_taxids = {}
        with open(conf.tax_childs, "r") as f:
            for line in f:
                try:
                    taxid, childs = line.strip()\
                        .split(',')
                except ValueError:
                    continue
                taxid2child_taxids[taxid] = set(childs.split(";"))
        return taxid2child_taxids

    def get_ranks(self, *taxids: str) -> Mapping[str, str]:
        result = self._ncbi.get_ranks(*taxids)
        return {str(k): v for k, v in result.items()}

    def get_names(self, *taxids: str) -> Mapping[str, str]:
        result = self._ncbi.get_names(*taxids)
        return {str(k): v for k, v in result.items()}

    def clade_lineage(self, clade: str) -> List[str]:
        return self.get_lineage(self.clade2taxid[clade])
        
    def get_lineage(self, taxid: str) -> List[str]:
        result = self._ncbi.get_lineage(taxid)
        return [str(x) for x in result]
    
    def write_clade_table(self, file: str, tree: Index):
        # get taxonomic data for all taxids
        all_taxids = set(self.clade2taxid.values())
        all_taxids.discard("unclassified")
        taxid2name = self.get_names(*all_taxids)
        taxid2rank = self.get_ranks(*all_taxids)
        # write to disk
        with open(file, "w") as f:
            f.write("clade,taxid,taxa,rank\n")
            f.write("unclassified,unclassified,unclassified,no rank\n")
            for node in tree.pool[1:]:
                node_name = tree.give_branch_name(node.name)
                taxid = self.clade2taxid[node.name]
                f.write("%s,%s,%s,%s\n" % (node_name, taxid, taxid2name[taxid], taxid2rank[taxid]))
